# ArcGIS Python script to translate NCGMP09-style DescriptionOfMapUnits table
# to fully-formatted USGS-style DMU text. Output is OpenOffice.org XML.
# OpenOffice.org is available for free and its FILE FORMAT IS OPEN: UNENCUMBERED
# BY INTELLECTUAL PROPERTY RESTRICTIONS and with an authoritative public description.
# Script developed and tested with OpenOffice.org version 3.1.1, ArcGIS version
# 9.3.1, and Python 2.5
#
# To produce an MSWord .doc text file, open output of this script in OOo, select
# File > Save As, then select format "Microsoft Word 97/2000/XP (.doc)". This file
# can then be dropped Ainto dobe InDesign with complete retention of all formatting.
#
#  Usage:
#    prompt> ncgmp09_dmu2odt.py <input_geodatabase> <output_OOo_text_file>
#
#  Ralph Haugerud, USGS Seattle, rhaugerud@usgs.gov

import sys, arcgisscripting, zipfile, os.path, shutil

versionString = 'ncgmp09_dmuOdt.py, version of 8 February 2010'

template = 'ncgmp09_dmu2odt-template.zip'

# constants for ODF XML
startContent = """<?xml version="1.0" encoding="UTF-8"?>
<office:document-content xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0"
xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0"
xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0"
xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0"
xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:dc="http://purl.org/dc/elements/1.1/"
xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0" xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
xmlns:math="http://www.w3.org/1998/Math/MathML" xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0"
xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0" xmlns:ooo="http://openoffice.org/2004/office"
xmlns:ooow="http://openoffice.org/2004/writer" xmlns:oooc="http://openoffice.org/2004/calc"
xmlns:dom="http://www.w3.org/2001/xml-events" xmlns:xforms="http://www.w3.org/2002/xforms"
xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xmlns:rpt="http://openoffice.org/2005/report" xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2"
xmlns:rdfa="http://docs.oasis-open.org/opendocument/meta/rdfa#"
xmlns:field="urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0" office:version="1.2"><office:scripts/>
<office:font-face-decls>
<style:font-face style:name="FGDCGeoAge" svg:font-family="FGDCGeoAge" style:font-family-generic="decorative" style:font-pitch="variable" style:font-charset="x-symbol"/>
<style:font-face style:name="Tahoma1" svg:font-family="Tahoma"/><style:font-face style:name="Times New Roman" svg:font-family="&apos;Times New Roman&apos;" style:font-family-generic="roman" style:font-pitch="variable"/>
<style:font-face style:name="Arial" svg:font-family="Arial" style:font-family-generic="swiss" style:font-pitch="variable"/><style:font-face style:name="Andale Sans UI" svg:font-family="&apos;Andale Sans UI&apos;" style:font-family-generic="system" style:font-pitch="variable"/>
<style:font-face style:name="Tahoma" svg:font-family="Tahoma" style:font-family-generic="system" style:font-pitch="variable"/></office:font-face-decls>
<office:automatic-styles>
<style:style style:name="P1" style:family="paragraph" style:parent-style-name="Standard"><style:paragraph-properties fo:line-height="0.1701in"/></style:style>
<style:style style:name="T1" style:family="text"><style:text-properties fo:font-weight="bold" style:font-weight-asian="bold" style:font-weight-complex="bold"/></style:style>
<style:style style:name="T2" style:family="text"><style:text-properties style:text-position="super 58%"/></style:style>
<style:style style:name="T3" style:family="text"><style:text-properties style:text-position="super 58%" fo:font-size="10pt" style:font-size-asian="10pt" style:font-size-complex="10pt"/></style:style>
<style:style style:name="T4" style:family="text"><style:text-properties fo:font-size="10pt" style:font-size-asian="10pt" style:font-size-complex="10pt"/></style:style>
<style:style style:name="T5" style:family="text"><style:text-properties style:font-name="FGDCGeoAge"/></style:style>
<style:style style:name="T6" style:family="text"><style:text-properties style:font-name="FGDCGeoAge" fo:font-size="9pt" style:font-size-asian="9pt" style:font-size-complex="9pt"/></style:style>
<style:style style:name="T7" style:family="text"><style:text-properties style:text-position="sub 58%"/></style:style>
<style:style style:name="T8" style:family="text"><style:text-properties fo:font-style="italic" style:font-style-asian="italic" style:font-style-complex="italic"/></style:style>
</office:automatic-styles>
<office:body><office:text><text:sequence-decls>
<text:sequence-decl text:display-outline-level="0" text:name="Illustration"/>
<text:sequence-decl text:display-outline-level="0" text:name="Table"/>
<text:sequence-decl text:display-outline-level="0" text:name="Text"/>
<text:sequence-decl text:display-outline-level="0" text:name="Drawing"/>
</text:sequence-decls>
"""
endContent =    '</office:text></office:body></office:document-content>'
## paragraph styles
startH1 =       '<text:p text:style-name="Heading_20_1st">'
startH2 =       '<text:p text:style-name="Heading_20_2nd">'
startH3 =       '<text:p text:style-name="Heading_20_3rd">'
startH4 =       '<text:p text:style-name="Heading_20_4th">'
startH5 =       '<text:p text:style-name="Heading_20_5th">'
startHNote =    '<text:p text:style-name="Headnote">'
startDMU1a =    '<text:p text:style-name="DMU1-first">'
startDMU1 =     '<text:p text:style-name="DMU1">'
startDMU2 =     '<text:p text:style-name="DMU2">'
startDMU3 =     '<text:p text:style-name="DMU3">'
startDMU4 =     '<text:p text:style-name="DMU4">'
startDMU5 =     '<text:p text:style-name="DMU5">'
startDMUPara =  '<text:p text:style-name="DMU_20_paragraph">'
endPara =       '</text:p>'
## character styles
startBold =     '<text:span text:style-name="T1">'
startItalic =   '<text:span text:style-name="T8">'
startSuper =    '<text:span text:style-name="T2">'
startSub =      '<text:span text:style-name="T7">'
startGeoAge =   '<text:span text:style-name="T6">'
endChar =       '</text:span>'
## other constants
tab =           '<text:tab/>'
emDash = '&#8212;'

styleDict =    {'DMU1':startDMU1,
                'DMU2':startDMU2,
                'DMU3':startDMU3,
                'DMU4':startDMU4,
                'DMU5':startDMU5,
                'Heading2':startH2,
                'Heading3':startH3,
                'Heading4':startH4,
                'Heading5':startH5,
                'HeadNote':startHNote }

usage = """Usage:
   systemprompt> ncgmp09_dmuOdt.py <geodatabaseName> <outputName>
   <geodatabaseName> is an existing NCGMP09-style database, with suffix
   <outputName> is the name (should end in .odt) of a file that does NOT exist

"""
    

def addMsgAndPrint(gp,msg, severity=0): 
    # prints msg to screen and adds msg to the geoprocessor (in case this is run as a tool) 
    # print msg 
    try: 
        for string in msg.split('\n'): 
            # Add appropriate geoprocessing message 
            if severity == 0: 
                gp.AddMessage(string) 
            elif severity == 1: 
                gp.AddWarning(string) 
            elif severity == 2: 
                gp.AddError(string) 
    except: 
        pass

# dictionary of supported markup tags and equivalent OpenOffice XML
tagDict = {'<i>':startItalic,   '</i>':endChar,
           '<b>':startBold,     '</b>':endChar,
           '<sub>':startSub,    '</sub>':endChar,
           '<sup>':startSuper,    '</sup>':endChar,
           '<ga>':startGeoAge,  '</ga>':endChar,
           '<p>':'|',   '</p>':'', '<br>':'|'}

# build dictionary of temporary markup tags
markupTempDict = {}
for key in tagDict.keys():
    tKey = key.replace('<','||')
    tKey = tKey.replace('>','|||')
    markupTempDict[key] = tKey

# replace < and > (inequality) symbols without corrupting markup tags
#   do so by replacing markup tags with pipe-delimited temporary markup tags
def markupInequalities(desc1):
    for key in markupTempDict:
        desc1 = desc1.replace(key,markupTempDict[key])
    desc1 = desc1.replace('<','&lt;')
    desc1 = desc1.replace('>','&gt;')
    for key in markupTempDict:
        desc1 = desc1.replace(markupTempDict[key],key)
    return desc1

# replace markup tags with OOo markup, then split into paragraphs
def unravel(desc1):
    desc1 = markupInequalities(desc1)
    desc3 = desc1.split('|')
    #print 'length desc3 = ',len(desc3)
    return desc3

def buildDmuDict(geoDB):
    # get DescriptionOfMapUnits rows, sorted by HierarchyKey
    print '  Building dictionary of DMU rows'
    gp.workspace = geoDB
    rows = gp.searchCursor('DescriptionOfMapUnits',"","","","HierarchyKey A")
    # process rows
    row = rows.next()
    hKeys = []
    dmuDict = {}
    while row:
        hKey = row.GetValue('HierarchyKey')
        hKeys.append(hKey)
        pStyle = row.GetValue('ParagraphStyle')
        name = row.GetValue('Name')
        age = row.GetValue('Age')
        if age == None: age = ''
        label = row.GetValue('Label')
        if label == None: label = ''
        #print hKey,name,age,label
        description = row.GetValue('Description')
        dmuDict[hKey] = [pStyle,label,name,age,description]
        row = rows.next()
    return hKeys,dmuDict 

def parentAge(hKey,dmuDict):
    parent = hKey[0:hKey.rfind('-')]
    return dmuDict[parent][3]

def checkHierarchy(hKeys,dmuDict):
    # is parent of higher rank?
    # are all siblings of same rank?
    # are there gaps in the sibling cohort?
    # does parent of DMU(n) have style DMU(n-1)?
    # is parent of Heading(n) have style Heading(n-1)
    # if style = Headnote, is parent a Heading?
    # if style = N/A, is it child-less?
    hErrors = []
    print '  Checking DMU hierarchy'
    for hKey in dmuDict.keys():
        if hKey.find('-') <> -1:  # skip if at top of hierarchy
            pHKey = hKey[0:hKey.rfind('-')]
            style = dmuDict[hKey][0]
            pStyle = dmuDict[pHKey][0]
    return hErrors

def main(geoDB, outfl, codepath):
    #  make dictionary of DMU rows and build list of hKeys
    hKeys,dmuDict = buildDmuDict(geoDB)
    # check hierarchy for consistency
    hierarchyErrors = checkHierarchy(hKeys,dmuDict)
    # setup .odt file without content
    addMsgAndPrint(gp,'  Copying template to '+outfl)
    shutil.copyfile(codepath+'/'+template,outfl)
    # open output file content.xml
    contentXML = open('content.xml','w')
    contentXML.write(startContent+'\n')
    if dmuDict[hKeys[0]][2].upper() <> 'DESCRIPTION OF MAP UNITS':
        contentXML.write(startH1+'Description of Map Units'+endPara+'\n')
    # process rows, adding formatting and writing to outputfile
    print '  Processing DMU rows and writing content.xml'
    headingPrecedes = False
    for hKey in hKeys:
        dmuLine = dmuDict[hKey]
        pStyle = dmuLine[0]
        label = dmuLine[1]
        if label == None: label = ''
        name = dmuLine[2]
        age = dmuLine[3]
        if age == None: age = ''
        description = dmuLine[4]
        if pStyle <> 'N/A':
            pText = styleDict[pStyle]
            if pStyle.find('DMU') <> -1:
                if headingPrecedes and pStyle == 'DMU1':
                    pText = startDMU1a
                descParagraphs = unravel(description)
                pText = pText+tab+startGeoAge+label+endChar+tab+startBold+name
                if age <> '':
                    if age <> parentAge(hKey,dmuDict):
                        pText = pText + ' ('+age+')'
                pText = pText + endChar + emDash + descParagraphs[0]+endPara
                # add additional DMU paragraphs, if present
                for n in range(1,len(descParagraphs)):
                    pText = pText+startDMUPara+descParagraphs[n]+endPara
                    
                headingPrecedes = False
            else: #pText nc 'DMU'
                pText = pText+name+endPara
                if description <> None:
                    if description[0] <> '[': description = '['+description
                    if description[-1:] <> ']': description = description + ']'
                    pText = pText + startHNote + description + endPara
                headingPrecedes = True
            # add OOo tags to label, if necessary
            for key in tagDict:
                pText = pText.replace(key,tagDict[key])
                pText = pText.replace(key.upper(),tagDict[key])
            try:
                contentXML.write(pText.encode("utf-8") + '\n')
            except:
                print 'Could not write '+name+' '+hKey
                print description.encode('utf-8')
    contentXML.write(endContent+'\n')
    contentXML.close()
    # add outputfile to zip archive outfl (xxx.odt)
    addMsgAndPrint(gp,'  Zipping content.xml into '+outfl)
    odtFile = zipfile.ZipFile(outfl,'a')
    odtFile.write('content.xml')
    odtFile.close()


if len(sys.argv) == 3 and os.path.exists(sys.argv[1]) and not os.path.exists(sys.argv[2]):
    try:
        print '  Creating geoprocessing object'
        gp = arcgisscripting.create(9.3)       
    except:
	addMsgAndPrint(gp,'Could not create geoprocessing object--no ArcGIS license?')
    else:
        if sys.argv[2][-4:] == '.odt':
            outfl = sys.argv[2]
        else:
            outfl = sys.argv[2]+'.odt'
        main(sys.argv[1], outfl, os.path.dirname(sys.argv[0]))
else:
    print usage

